@@ -23,9 +23,9 @@
 from __future__ import with_statement
 
 import commands
+import datetime
 import errno
 import logging
-import os
 import optparse
 import pprint
 import signal
@@ -42,6 +42,7 @@ except ImportError:
 
 import eventlet
 import eventlet.backdoor
+from eventlet.green import os
 
 import spawning
 import spawning.util
@@ -81,13 +82,20 @@ def environ():
 
     env['PYTHONPATH'] = ':'.join(new_path)
     return env
+    
+class Child(object):
+    def __init__(self, pid, kill_pipe):
+        self.pid = pid
+        self.kill_pipe = kill_pipe
+        self.active = True
+        self.forked_at = datetime.datetime.now()
 
 class Controller(object):
     sock = None
     factory = None
     args = None
     config = None
-    child_pipes = None
+    children = None
     keep_going = True
     panic = False
     log = None
@@ -99,13 +107,14 @@ class Controller(object):
         self.factory = factory
         self.config = spawning.util.named(factory)(args)
         self.args = args
-        self.child_pipes = {}
+        self.children = {}
         self.log = logging.getLogger('Spawning')
         if not kwargs.get('log_handler'):
             self.log.addHandler(logging.StreamHandler())
         self.log.setLevel(logging.DEBUG)
         self.controller_pid = os.getpid()
         self.num_processes = int(self.config.get('num_processes', 0))
+        self.started_at = datetime.datetime.now()
 
     def spawn_children(self, number=1):
         parent_pid = os.getpid()
@@ -119,7 +128,7 @@ class Controller(object):
                 print_exc('Could not fork child! Panic!')
                 ### TODO: restart
 
-            if not child_pid:
+            if not child_pid:      # child process
                 os.close(parent_side)
                 command = [sys.executable, '-c',
                     'import sys; from spawning import spawning_child; spawning_child.main()',
@@ -131,22 +140,30 @@ class Controller(object):
                 if self.args['reload'] == 'dev':
                     command.append('--reload')
                 env = environ()
-                env['EVENTLET_THREADPOOL_SIZE'] = str(self.config.get('threadpool_workers', 0))
+                tpool_size = int(self.config.get('threadpool_workers', 0))
+                assert tpool_size >= 0, (tpool_size, 'Cannot have a negative --threads argument')
+                if not tpool_size in (0, 1):
+                    env['EVENTLET_THREADPOOL_SIZE'] = str(tpool_size)
                 os.execve(sys.executable, command, env)
+                
+            # controller process
             os.close(child_side)
-            self.child_pipes[child_pid] = parent_side
-
+            self.children[child_pid] = Child(child_pid, parent_side)
+    
+    def children_count(self):
+        return len(self.children)
+    
     def runloop(self):
         while self.keep_going:
             eventlet.sleep(0.1)
             ## Only start the number of children we need
-            number = self.num_processes - len(self.child_pipes)
+            number = self.num_processes - self.children_count()
             if number > 0:
                 self.log.debug('Should start %d new children', number)
                 self.spawn_children(number=number)
                 continue
 
-            if not self.child_pipes:
+            if not self.children:
                 ## If we don't yet have children, let's loop
                 continue
 
@@ -157,9 +174,10 @@ class Controller(object):
                 if e.errno != errno.EINTR:
                     raise
 
-            if pid and self.child_pipes.get(pid):
+            if pid and self.children.get(pid):
                 try:
-                    os.close(self.child_pipes.pop(pid))
+                    child = self.children.pop(pid)
+                    os.close(child.kill_pipe)
                 except (IOError, OSError):
                     pass
 
@@ -172,21 +190,25 @@ class Controller(object):
     def handle_sighup(self, *args, **kwargs):
         ''' Pass `no_restart` to prevent restarting the run loop '''
         self.kill_children()
+        self.spawn_children(number=self.num_processes)
+        # TODO: nothing seems to use no_restart, can it be removed?
         if not kwargs.get('no_restart', True):
             self.runloop()
 
     def kill_children(self):
-        for pid, pipe in self.child_pipes.items():
+        for pid, child in self.children.items():
             try:
-                os.write(pipe, ' ')
-                os.close(pipe)
+                os.write(child.kill_pipe, 'k')
+                child.active = False
+                # all maintenance of children's membership happens in runloop()
+                # as children die and os.wait() gets results
             except OSError, e:
                 if e.errno != errno.EPIPE:
                     raise
 
     def handle_deadlychild(self, *args, **kwargs):
         if self.keep_going:
-            self.log.debug('Child indicating it\'s about to die, starting a replacement')
+            self.log.debug('A child intends to die, starting replacement before it dies')
             self.spawn_children(number=1)
 
     def run(self):
@@ -204,6 +226,12 @@ class Controller(object):
 
         signal.signal(signal.SIGHUP, self.handle_sighup)
         signal.signal(signal.SIGUSR1, self.handle_deadlychild)
+        
+        if self.config.get('status_port'):
+            from spawning.util import status
+            eventlet.spawn(status.Server, self, 
+                self.config['status_host'], self.config['status_port'])
+            
         try:
             self.runloop()
         except KeyboardInterrupt:
@@ -307,6 +335,9 @@ def main():
         help='If given, gather coverage data from the running program and make the '
             'coverage report available from the /_coverage url. See the figleaf docs '
             'for more info: http://darcs.idyll.org/~t/projects/figleaf/doc/')
+    parser.add_option('--sysinfo', dest='sysinfo', action='store_true',
+        help='If given, gather system information data and make the '
+            'report available from the /_sysinfo url.')
     parser.add_option('-m', '--max-memory', dest='max_memory', type='int', default=0,
         help='If given, the maximum amount of memory this instance of Spawning '
             'is allowed to use. If all of the processes started by this Spawning controller '
@@ -316,12 +347,16 @@ def main():
             help='Start a backdoor bound to localhost:3000')
     parser.add_option('-a', '--max-age', dest='max_age', type='int',
         help='If given, the maximum amount of time (in seconds) an instance of spawning_child '
-            'is allowed to run. Once this time limit has expired a SIGHUP will be sent to '
-            'spawning_controller, causing it to restart all of the child processes.')
+            'is allowed to run. Once this time limit has expired the child will'
+            'gracefully kill itself while the server starts a replacement.')
     parser.add_option('--no-keepalive', dest='no_keepalive', action='store_true',
             help='Disable HTTP/1.1 KeepAlive')
     parser.add_option('-z', '--z-restart-args', dest='restart_args',
         help='For internal use only')
+    parser.add_option('--status-port', dest='status_port', type='int', default=0,
+        help='If given, hosts a server status page at that port.  Two pages are served: a human-readable HTML version at http://host:status_port/status, and a machine-readable version at http://host:status_port/status.json')
+    parser.add_option('--status-host', dest='status_host', type='string', default='',
+        help='If given, binds the server status page to the specified local ip address.  Defaults to the same value as --host.  If --status-port is not supplied, the status page will not be activated.')
 
     options, positional_args = parser.parse_args()
 
@@ -437,6 +472,11 @@ def main():
     # If you tell me to watch something, I'm going to reload then
     if options.watch:
         options.reload = True
+        
+    if options.status_port == options.port:
+        options.status_port = None
+        sys.stderr.write('**> Status port cannot be the same as the service port, disabling status.\n')
+        
 
     factory_args = {
         'verbose': options.verbose,
@@ -450,10 +490,13 @@ def main():
         'access_log_file': options.access_log_file,
         'pidfile': options.pidfile,
         'coverage': options.coverage,
+        'sysinfo': options.sysinfo,
         'no_keepalive' : options.no_keepalive,
         'max_age' : options.max_age,
         'argv_str': " ".join(sys.argv[1:]),
         'args': positional_args,
+        'status_port': options.status_port,
+        'status_host': options.status_host or options.host
     }
     start_controller(sock, factory, factory_args)
 
